#include "octutils.h"
#include "channelflow/periodicfunc.h"


using namespace std;

namespace channelflow {

void save(const std::string& filebase, Real  T, const FieldSymmetry& tau) {
  string filename = appendSuffix(filebase, ".asc");
  ofstream os(filename.c_str());
  const int p = 16;  // digits of precision
  const int w = p+7; // width of field
  os << setprecision(p) << scientific;
  os << setw(w) << T << " %T\n";
  os << setw(w) << tau.su() << " %s\n";
  os << setw(w) << tau.sx() << " %sx\n";
  os << setw(w) << tau.sy() << " %sy\n";
  os << setw(w) << tau.sz() << " %sz\n";
  os << setw(w) << tau.ax() << " %ax\n";
  os << setw(w) << tau.az() << " %az\n";
};

void load(const std::string& filebase, Real& T, FieldSymmetry& tau) {
  string filename = appendSuffix(filebase, ".asc");
  ifstream is(filename.c_str());
  string junk;
  int su, sx, sy, sz;
  Real ax, az;
  bool error = false;
  is >> T >> junk;
  if (junk != "%T") error = true;
  is >> su >> junk;
  if (junk != "%s" || abs(su) != 1) error = true;
  is >> sx >> junk;
  if (junk != "%sx" || abs(sx) != 1) error = true;
  is >> sy >> junk;
  if (junk != "%sy" || abs(sy) != 1) error = true;
  is >> sz >> junk;
  if (junk != "%sz" || abs(sz) != 1) error = true;
  is >> ax >> junk;
  if (junk != "%ax" || abs(ax) > 0.5) error = true;
  is >> az >> junk;
  if (junk != "%az" || abs(az) > 0.5) error = true;

  if (error) {
    cout << "Error in loading T,su,sx,sy,sz,ax,az from file " << filename << endl;
    cout << "Exiting." << endl;
    exit(1);
  }
  tau = FieldSymmetry(sx,sy,sz,ax,az,su);
}

void plotfield(const FlowField& u_, const string& outdir, const string& label,
	       int xstride, int ystride, int zstride) {

  string preface = outdir;
  if (label.length() > 0)
    preface += label + "_";

  mkdir(outdir);

  FlowField& u = (FlowField&) u_;
  fieldstate xs = u.xzstate();
  fieldstate ys = u.ystate();

  const int Ny2 = (u.Ny()-1)/2;
  //Vector x = periodicpoints(Nx, u.Lx());
  //Vector y = chebypoints(Ny, u.a(), u.b());
  //Vector z = periodicpoints(Nz, u.Lz());
  //x.save(outdir + "x");
  //y.save(outdir + "y");
  //z.save(outdir + "z");

  u.makePhysical();
  u.saveSlice(0, 0, 0,   preface + "u_yz");
  u.saveSlice(0, 1, 0,   preface + "v_yz");
  u.saveSlice(0, 2, 0,   preface + "w_yz");
  u.saveSlice(1, 0, Ny2, preface + "u_xz");
  u.saveSlice(1, 1, Ny2, preface + "v_xz");
  u.saveSlice(1, 2, Ny2, preface + "w_xz");
  u.saveSlice(1, 0, 11, preface + "u_xz_11");
  u.saveSlice(1, 1, 11, preface + "v_xz_11");
  u.saveSlice(1, 2, 11, preface + "w_xz_11");
  u.saveSlice(1, 0, 23, preface + "u_xz_23");
  u.saveSlice(1, 1, 23, preface + "v_xz_23");
  u.saveSlice(1, 2, 23, preface + "w_xz_23");
  u.saveSlice(2, 0, 0,   preface + "u_xy");
  u.saveSlice(2, 1, 0,   preface + "v_xy");
  u.saveSlice(2, 2, 0,   preface + "w_xy");
  u.makeSpectral();


  //FlowField vort = curl(u);
  //vort.makePhysical();
  //vort.saveSlice(0, 0, 0, preface + "vort_yz");

  //FlowField drag = ydiff(u);
  //drag.makePhysical();
  //drag.saveSlice(1, 0, Ny-1, preface + "drag_xz");

  //ChebyCoeff U = Re(u.profile(0,0,0));
  //U.makePhysical();
  //U.save("U");

  u.makeState(xs,ys);
}

// Produce plots of various slices, etc.
void plotxavg(const FlowField& u_, const std::string& outdir,
	      const std::string& label) {

  FlowField& u = (FlowField&) u_;
  fieldstate xs = u.xzstate();
  fieldstate ys = u.ystate();
  u.makeSpectral();

  FlowField uavg(4, u.Ny(), u.Nz(), u.Nd(), u.Lx(), u.Lz(), u.a(), u.b(), Spectral, Spectral);
  for (int i=0; i<u.Nd(); ++i)
    for (int my=0; my<u.My(); ++my)
      for (int mz=0; mz<u.Mz(); ++mz)
	uavg.cmplx(0,my,mz,i) = u.cmplx(0,my,mz,i);

  mkdir(outdir);
  string preface = outdir;
  if (label.length() > 0)
    preface += label + "_";

  uavg.makePhysical();
  uavg.saveSlice(0, 0, 0,   preface + "u_yz_xavg");
  uavg.saveSlice(0, 1, 0,   preface + "v_yz_xavg");
  uavg.saveSlice(0, 2, 0,   preface + "w_yz_xavg");

  u.makeState(xs,ys);
}


void plotspectra(const FlowField& u_, const string& outdir, const string& label,
		 bool showpadding) {

  FlowField& u = (FlowField&) u_;
  fieldstate xs = u.xzstate();
  fieldstate ys = u.ystate();

  string preface = outdir;
  if (label.length() > 0)
    preface += label + "_";

  mkdir(outdir);
  u.makeSpectral();
  u.saveSpectrum(preface + "xzspec", true, showpadding);

  u.makePhysical_y();
  ofstream ucs((preface + "yspec.asc").c_str());

  ucs << "% sum of abs vals of Cheby coeffs of components u_{kx,kz}(y) " << endl;

  ChebyTransform trans(u.Ny());

  const int kxmax = lesser(u.kxmax(), 8);
  const int kzmax = lesser(u.kzmax(), 8);
  const int dk = 1;

  for (int kx=0; kx<=kxmax; kx += dk) {
    //for (int kz=0; kz<=kzmax; kz += dk) {
    int kz=lesser(kx,kzmax);
    BasisFunc prof = u.profile(u.mx(kx),u.mz(kz));
      prof.makeSpectral(trans);
      ucs << "% kx,kz == " << kx << ", " << kz << endl;
      for (int ny=0; ny<prof.Ny(); ++ny) {
	Real sum = 0.0;
	for (int i=0; i<prof.Nd(); ++i)
	  sum += abs(prof[i][ny]);
	ucs << sum << ' ';
      }
      ucs << '\n';
  }

  u.makeState(xs,ys);

}

/*****************
TimeStepParams::TimeStepParams(Real dt_, Real dtmin_, Real dtmax_,
			       Real CFLmin_, Real CFLmax_, bool variable_)
  :
  dt(dt_),
  dtmin(dtmin_),
  dtmax(dtmax_),
  CFLmin(CFLmin_),
  CFLmax(CFLmax_),
  variable(variable_)
{}


void params2timestep(const TimeStepParams& p, Real T, int& N, TimeStep& dt) {

  // Set CFL adjustment time dT to an integer division of T near 1
  N = Greater(iround(T), 1);
  const Real dT = T/N;
  if (p.variable)
    dt = TimeStep(p.dt, p.dtmin, p.dtmax, dT, p.CFLmin, p.CFLmax);
  else
    dt = TimeStep(p.dt, p.dt, p.dt, dT, p.CFLmin, p.CFLmax);
}
******************/

ArgList::ArgList()
  :
  args_(),
  used_(),
  helpmode_(false),
  errormode_(false)
{}


ArgList::ArgList(int argc, char* argv[], const string& purpose)
  :
  args_(argc),
  used_(argc),
  helpmode_(false),
  errormode_(false)
{
  string h0("-h");
  string h1("--help");
  for (int i=0; i<argc; ++i) {
    args_[i] = string(argv[i]);
    if (args_[i] == h0 || args_[i] == h1) {
      helpmode_ = true;
      used_[i] = true;
    }
    else
      used_[i] = false;
  }
  if (helpmode_) {
    cerr << argv[0] << " : \n\t"
	 << purpose << endl << endl;
    if (argc > 1)
      cerr << "options : " << endl;
  }
  used_[0] = true;
}


bool ArgList::helpmode() const {
  return helpmode_;
}

bool ArgList::errormode() const {
  return errormode_;
}

int ArgList::remaining() const {
  int unused = 0;
  for (int i=0; i<used_.length(); ++i)
    if (!used_[i])
      ++unused;
  return unused;
}

// The magic setw constants are a quick and dirty way to line up the columns
// nicely as long as the strings aren't too long. If you want to reimplement
// formatting more intelligently, please do!
void ArgList::printhelp(int position, const std::string& name,
			const string& helpstr) {
  cerr.setf(ios::left);
  cerr << "  ";
  cerr << setw(17) << name;
  cerr << setw(48) << string("(trailing arg " + i2s(position) + ")");
  cerr.unsetf(ios::left);
  cerr << helpstr << endl;
}

void ArgList::printhelp(const string& sopt, const string& lopt,
			const string& type, const string& defalt,
			const string& helpstr) {
  cerr.setf(ios::left);
  cerr << "  " << setw(8) << sopt << "  ";
  cerr << setw(20) << lopt
       << setw(10) << type;

  if (defalt.length() != 0)
    cerr << "  default == "
	 << setw(12) << defalt;
  else
    cerr << setw(25) << "";
  cerr.unsetf(ios::left);
  cerr << helpstr << endl;
}

bool ArgList::getflag(const string& sopt, const string& lopt,
		      const string& helpstr) {
  if (helpmode_) {
    printhelp(sopt, lopt, "", "", helpstr);
    return false;
  }

  bool b = false;
  int N = args_.length();
  for (int n=N-1; n>=1; --n) {
    if (args_[n] == sopt || args_[n] == lopt) {
      b = true;
      used_[n] = true;
      break;
    }
  }
  return b;
}



bool ArgList::getbool(const string& sopt, const string& lopt,
		      const string& helpstr) {
  bool b = false;

  int N = args_.length();
  bool found = false;
  for (int n=N-1; n>=1; --n) {
    if (args_[n] == sopt || args_[n] == lopt) {
      found = true;
      if (n<N-1 && args_[n+1] == "true")
	b = true;
      else if (n<N-1 && args_[n+1] == "false")
	b = false;
      else {
	cerr << "error : option " << sopt << " or " << lopt
	     << " should be followed by 'true' or 'false'." << endl;
	exit(1);
      }
      used_[n]   = true;
      used_[n+1] = true;
      break;
    }
  }
  if (helpmode_) {
    printhelp(sopt, lopt, "<bool>", "", helpstr);
    return b;
  }
  else if (!found) {
    helpmode_ = true;
    errormode_ = true;
    cerr << "Missing required argument(s). Arguments are : " << endl;
    printhelp(sopt, lopt, "<bool>", "", helpstr);
    //exit(1);
  }
  return b;
}

int  ArgList::getint (const string& sopt, const string& lopt,
		      const string& helpstr) {
  int i=0;
  bool found = false;
  int N = args_.length();
  for (int n=N-1; n>=1; --n) {
    if (args_[n] == sopt || args_[n] == lopt) {
      found = true;
      if (n<N-1)
	i = atoi(args_[n+1].c_str());
      else {
	cerr << "error : option " << sopt << " or " << lopt
	     << " should be followed by an integer." << endl;
	exit(1);
      }
      used_[n]   = true;
      used_[n+1] = true;
      break;
    }
  }
  if (helpmode_) {
    printhelp(sopt, lopt, "<int>", "", helpstr);
    return i;
  }
  else if (!found) {
    helpmode_ = true;
    errormode_ = true;
    cerr << "Missing required argument(s). Arguments are : " << endl;
    printhelp(sopt, lopt, "<int>", "", helpstr);
    //exit(1);
  }
  return i;
}

Real ArgList::getreal(const string& sopt, const string& lopt,
		      const string& helpstr) {
  Real r=0.0;

  int N = args_.length();
  bool found = false;
  for (int n=N-1; n>=1; --n) {
    if (args_[n] == sopt || args_[n] == lopt) {
      found = true;
      if (n<N-1)
	r = atof(args_[n+1].c_str());
      else {
	cerr << "error : option " << sopt << " or " << lopt
	     << " should be followed by a real number." << endl;
	exit(1);
      }
      used_[n]   = true;
      used_[n+1] = true;
      break;
    }
  }
  if (helpmode_) {

    printhelp(sopt, lopt, "<real>", "", helpstr);
    return r;
  }
  else if (!found) {
    helpmode_ = true;
    errormode_ = true;
    cerr << "Missing required argument(s). Arguments are : " << endl;
    printhelp(sopt, lopt, "<real>", "", helpstr);
    //exit(1);
  }
  return r;
}

string ArgList::getpath(const string& sopt, const string& lopt,
			const string& helpstr) {
  return pathfix(getstr(sopt,lopt,helpstr));
}

string ArgList::getstr(const string& sopt, const string& lopt,
		       const string& helpstr) {
  string s("");

  bool found = false;
  int N = args_.length();
  for (int n=N-1; n>=1; --n) {
    if (args_[n] == sopt || args_[n] == lopt) {
      found = true;
      if (n<N-1)
	s = string(args_[n+1]);
      else {
	cerr << "error : option " << sopt << " or " << lopt
	     << " should be followed by a string." << endl;
	exit(1);
      }
      used_[n]   = true;
      used_[n+1] = true;
      break;
    }
  }
  if (helpmode_) {
    printhelp(sopt, lopt, "<string>", "", helpstr);
    return s;
  }
  else if (!found) {
    helpmode_ = true;
    errormode_ = true;
    cerr << "Missing required argument(s). Arguments are : " << endl;
    printhelp(sopt, lopt, "<string>", "", helpstr);
    //exit(1);
  }
  return s;
}

bool ArgList::getbool(const string& sopt, const string& lopt, bool defalt,
		      const string& helpstr) {
  bool b = defalt;
  if (helpmode_) {
    printhelp(sopt, lopt, "<bool>", string(b ? "true" : "false"), helpstr);
    return b;
  }

  int N = args_.length();
  for (int n=N-1; n>=1; --n) {
    if (args_[n] == sopt || args_[n] == lopt) {
      if (n<N-1 && args_[n+1] == "true")
	b = true;
      else if (n<N-1 && args_[n+1] == "false")
	b = false;
      else {
	cerr << "error : option " << sopt << " or " << lopt
	     << " should be followed by 'true' or 'false'." << endl;
	exit(1);
      }
      used_[n]   = true;
      used_[n+1] = true;
      break;
    }
  }
  return b;
}

int  ArgList::getint (const string& sopt, const string& lopt, int defalt,
		      const string& helpstr) {

  int i=defalt;
  if (helpmode_) {
    printhelp(sopt, lopt, "<int>", i2s(defalt), helpstr);
    return i;
  }
  int N = args_.length();
  for (int n=N-1; n>=1; --n) {
    if (args_[n] == sopt || args_[n] == lopt) {
      if (n<N-1)
	i = atoi(args_[n+1].c_str());
      else {
	cerr << "error : option " << sopt << " or " << lopt
	     << " should be followed by an integer." << endl;
	exit(1);
      }
      used_[n]   = true;
      used_[n+1] = true;
      break;
    }
  }
  return i;
}

Real ArgList::getreal(const string& sopt, const string& lopt, Real defalt,
		      const string& helpstr) {
  Real r=defalt;
  if (helpmode_) {
    printhelp(sopt, lopt, "<real>", r2s(defalt), helpstr);
    return r;
  }

  int N = args_.length();
  for (int n=N-1; n>=1; --n) {
    if (args_[n] == sopt || args_[n] == lopt) {
      if (n<N-1)
	r = atof(args_[n+1].c_str());
      else {
	cerr << "error : option " << sopt << " or " << lopt
	     << " should be followed by a real number." << endl;
	exit(1);
      }
      used_[n]   = true;
      used_[n+1] = true;
      break;
    }
  }
  return r;
}

string ArgList::getpath(const string& sopt, const string& lopt,
			const string& defalt, const string& helpstr) {
  return pathfix(getstr(sopt,lopt,defalt,helpstr));
}

string ArgList::getstr(const string& sopt, const string& lopt,
		       const string& defalt, const string& helpstr) {
  string s(defalt);
  if (helpmode_) {
    printhelp(sopt, lopt, "<string>", defalt, helpstr);
    return defalt;
  }
  int N = args_.length();
  for (int n=N-1; n>=1; --n) {
    if (args_[n] == sopt || args_[n] == lopt) {
      if (n<N-1)
	s = string(args_[n+1]);
      else {
	cerr << "error : option " << sopt << " or " << lopt
	     << " should be followed by a string." << endl;
	exit(1);
      }
      used_[n]   = true;
      used_[n+1] = true;
      break;
    }
  }
  return s;
}

string ArgList::getstr(int position, const std::string& name,
		       const string& helpstr) {
  if (helpmode_) {
    printhelp(position, name, helpstr);
    return string();
  }

  int n = args_.length()-position;
  if (n<1 || used_[n]) {
    cerr << "error : " << name << " is required as the " << position
	 << "th trailing argument (counting backwards from end of arglist)"
	 << endl;;
    exit(1);
  }
  if (args_[n][0] == '-') {
    cerr << "error : should have a filename for " << position << "th trailing argument, ";
    cerr << "not option " << args_[n] << endl;
    exit(1);
  }
  used_[n] = true;
  return args_[n];
}

Real ArgList::getreal(int position, const std::string& name,
		      const string& helpstr) {
  if (helpmode_) {
    printhelp(position, name, helpstr);
    return 0.0;
  }

  int n = args_.length()-position;
  if (n<1 || used_[n]) {
    cerr << "error : " << name << " is required as the " << position
	 << "th trailing argument (counting backwards from end of arglist)"
	 << endl;;
    exit(1);
  }
  /***********************8
  if (args_[n][0] == '-') {
    cerr << "error : should have a Real for " << position << "th trailing argument, ";
    cerr << "not option " << args_[n] << endl;
    exit(1);
  }
  ******************************/
  used_[n] = true;
  return atof(args_[n].c_str());
}


void ArgList::check() {
  for (int n=0; n<used_.length(); ++n) {
    if (!used_[n]) {
      cerr << "error : unrecognized/repeated option/value " << args_[n] << endl;
      errormode_ = true;
    }
  }
  if (errormode_)
    exit(1);
  else if (helpmode_)
    exit(0);
}

void ArgList::save(const string& outdir) const {
  // Look for output directory and save command-line to outdir/argv[0].args
  string arg0 = args_[0];
  string exec;
  int slashpos = arg0.find_last_of('/');
  if (slashpos == -1) {
    exec = arg0;
  }
  else {
    exec = string(arg0, slashpos+1, arg0.length());
  }
  int extpos = exec.find(".x");
  if (extpos == -1)
    extpos = exec.find(".dx");
  if (extpos == -1)
    extpos = exec.length();

  string arse = string(exec, 0, extpos) + ".args";
  string afile = outdir + arse;
  ::mkdir(outdir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
  ofstream as(afile.c_str(), ios::app);
  for (int n=0; n<args_.length(); ++n)
    as << args_[n] << ' ';
  as << endl;
}

string pathfix(const string& path) {
  string rtn = path;
  if (rtn.length() > 0 && rtn[rtn.length()-1] != '/')
    rtn += "/";
  return rtn;
}

void mkdir(const string& dirname) {
  ::mkdir(dirname.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
}

Real PuFraction(const FlowField& u, const FieldSymmetry& s, int sign) {
  FlowField Pu = s(u);
  if (sign == 1)
    Pu += u;
  else
    Pu -= u;
  return 0.25*L2Norm2(Pu)/L2Norm2(u); // 0.25 is square of 1/2 from Pu = 1/2 (u+s(u))
}


// Return a translation that maximizes the s-symmetry of u.
// (i.e. return tau // that minimizes L2Norm(s(tau(u)) - tau(u))).
FieldSymmetry optimizePhase(const FlowField& u, const FieldSymmetry& s,
			    int Nsteps, Real residual, Real damp,
			    bool verbose, Real x0, Real z0) {

  FieldSymmetry tau;
  if (s.sx()==1 && s.sz()==1)
    ;
  else if ((s.sx() == -1) ^ (s.sz() == -1)) {

    // Newton search for x-translation that maximizes shift-rotate symmetry
    if (verbose)
      cout << "\noptimizePhase: Newton search on translation" <<endl;

    int i   = (s.sx() == -1) ? 0 : 2; // are we optimizing on x or z?
    Real L  = (s.sx() == -1) ? u.Lx() : u.Lz();
    array<Real> x(3);
    x[0] = x0;
    x[2] = z0;

    FlowField tu, stu, tux, stux, tuxx, stuxx;

    for (int n=0; n<Nsteps; ++n) {
      tau = FieldSymmetry(x[0]/u.Lx(), x[2]/u.Lz());

      tu  = tau(u);
      stu = s(tu);
      tux   = diff(tu, i, 1);
      stux  = diff(stu, i, 1);
      tuxx  = diff(tux, i, 1);
      stuxx = diff(stux, i, 1);

      // g(x) is df/dx where f == (L2Norm(stu - tu)) and t = tau(x/Lx,0)
      Real g    =  L2IP(stux, tu) - L2IP(stu, tux);
      Real dgdx = 2*L2IP(stux, tux) - L2IP(stuxx, tu) - L2IP(stu, tuxx);

      if (verbose) {
	Real f = L2Dist(stu, tu);
	cout << endl;
	cout << "Newton step n  == " << n << endl;
	cout << "tau.ax         == " << tau.ax() << endl;
	cout << "tau.az         == " << tau.az() << endl;
	cout << "zeroable g     == " << g << endl;
	cout << "L2Norm(stu-tu) == " << f << endl;
      }
      if (abs(g) < residual && verbose) {
	cout << "\noptimizePhase: Newton search converged." << endl;
	break;
      }

      Real dx = -g/dgdx;

      if (abs(damp*dx)/L > 0.5) {
	cout << "\nerror in optimizePhase(FlowField, FieldSymmetry) :" << endl;
	cout << "Newton search went beyond bounds. " << endl;
	cout << "     g == " << g << endl;
	cout << " dg/dx == " << dgdx << endl;
	cout << "    dx == " << damp*dx << endl;
	cout << "  L/2  == " << 0.5*L  << endl;
	cout << "Jostle by L/10 and try again." << endl;
	x[i] += L/10;
      }
      else
	x[i] += damp*dx;

    }
  }
  else {
    cout << "\nerror in optimizePhase(FlowField, FieldSymmetry) :" << endl;
    cout << "2d symmetrization is not yet implemented. Returning identity." << endl;
  }

  if (verbose) {
    FlowField tu = tau(u);
    FlowField stu = s(tu);
    cout << "\noptimal translation == " << tau << endl;
    cout << "Post-translation symmetry error == " << L2Dist(stu, tu) << endl;
  }
  return tau;
}

string t2s(Real t, bool inttime) {
  char buffer[12];
  if (inttime) {
    int it = iround(t);
    sprintf(buffer, "%d", it);
  }
  else
    sprintf(buffer, "%.3f", t);
  return string(buffer);
}

string fuzzyless(Real x, Real eps) {
  string rtn("false ");
  if (x < eps)
    rtn = "TRUE  ";
  else if (x < sqrt(eps))
    rtn = "APPROX";
  return rtn;
};

string pwd() {
  const int N=512;
  char* buff = new char[N];
  getcwd(buff, N);
  string rtn(buff);
  delete[] buff;
  return rtn;
}

string stub(const string& filename, const string& ext) {

  // Clip off path
  size_t s = filename.find_last_of("/");
  string f = (s == string::npos) ? filename : filename.substr(s+1, filename.length()-s);

  // Clip off extension
  size_t t = f.find(ext, f.length()-ext.length());
  string g = (t == string::npos) ? f : f.substr(0,t);

  return g;
}


// ======================================================================
// All Octave-dependent code goes within this ifdef
// #ifdef HAVE_OCTAVE

void setToZero(ColumnVector& x) {
  for (int i=0; i<x.length(); ++i)
    x(i) = 0.0;
}

void setToZero(Matrix& A) {
  for (int i=0; i<A.rows(); ++i)
    for (int j=0; j<A.cols(); ++j)
      A(i,j) = 0.0;
}

void save(const Matrix& A, const string& filebase) {
  string filename(filebase + ".asc");
  ofstream os(filename.c_str());
  os << setprecision(17);
  for (int i=0; i<A.rows(); ++i) {
    for (int j=0; j<A.cols(); ++j)
      os << A(i,j) << ' ';
    os << '\n';
  }
}

void print(const ColumnVector& x) {
  for (int i=0; i<x.length(); ++i)
    cout << x(i) << '\n';
}
void print(const Matrix& A) {
  for (int i=0; i<A.rows(); ++i) {
    for (int j=0; j<A.cols(); ++j)
      cout << A(i,j) << ' ';
    cout << '\n';
  }
}


void save(const ColumnVector& x, const string& filebase) {
  string filename(filebase + ".asc");
  ofstream os(filename.c_str());
  os << setprecision(17);
  for (int i=0; i<x.length(); ++i)
    os << x(i) << '\n';
}

void save(const ComplexColumnVector& x, const string& filebase) {
  string filename(filebase + ".asc");
  ofstream os(filename.c_str());
  os << setprecision(17);
  for (int i=0; i<x.length(); ++i)
    os << x(i).real() << ' ' << x(i).imag() << '\n';
}



Real L2Norm2(const ColumnVector& x) {
  Real sum=0.0;
  for (int i=0; i<x.length(); ++i)
    sum += x(i)*x(i);
  return sum;
}

Real L2Dist2(const ColumnVector& x,  const ColumnVector& y) {
  assert(x.length() == y.length());
  Real sum=0.0;
  for (int i=0; i<x.length(); ++i)
    sum += square(x(i)-y(i));
  return sum;
}
Real L2Norm(const ColumnVector& x) {
  return sqrt(L2Norm2(x));
}

Real L2Dist(const ColumnVector& x,  const ColumnVector& y) {
  return sqrt(L2Dist2(x,y));
}

void operator*=(ColumnVector& x, Real c) {
  for (int i=0; i<x.length(); ++i)
    x(i) *= c;
}

Real L2Norm2(const ComplexColumnVector& x) {
  Real sum=0.0;
  for (int i=0; i<x.length(); ++i)
    sum += abs2(x(i));
  return sum;
}

Real L2Dist2(const ComplexColumnVector& x,  const ComplexColumnVector& y) {
  assert(x.length() == y.length());
  Real sum=0.0;
  for (int i=0; i<x.length(); ++i)
    sum += abs2(x(i)-y(i));
  return sum;
}
Real L2Norm(const ComplexColumnVector& x) {
  return sqrt(L2Norm2(x));
}

Real L2Dist(const ComplexColumnVector& x,  const ComplexColumnVector& y) {
  return sqrt(L2Dist2(x,y));
}

void operator*=(ComplexColumnVector& x, Complex c) {
  for (int i=0; i<x.length(); ++i)
    x(i) *= c;
}


void fixDiri(ChebyCoeff& f) {
  Real fa = f.eval_a();
  Real fb = f.eval_b();
  Real mean = 0.5*(fb+fa);
  Real slop = 0.5*(fb-fa);
  f[0] -= mean;
  f[1] -= slop;
}

void fixDiriMean(ChebyCoeff& f) {
  Real fa = f.eval_a();
  Real fb = f.eval_b();
  Real fm = f.mean();
  f[0] -= 0.125*(fa+fb) + 0.75*fm;
  f[1] -= 0.5*(fb-fa);
  f[2] -= 0.375*(fa+fb) - 0.75*fm;
}

void fixDiriNeum(ChebyCoeff& f)  {

  Real ya = f.a();
  Real yb = f.b();
  f.setBounds(-1,1);
  Real a = f.eval_a();
  Real b = f.eval_b();
  Real c = f.slope_a();
  Real d = f.slope_b();

  // The numercial coeffs are inverse of the matrix (values found with Maple)
  // T0(-1)  T1(-1)  T2(-1)  T3(-1)     s0      a
  // T0(1)   T1(1)   T2(1)   T3(1)      s1      b
  // T0'(-1) T1'(-1) T2'(-1) T3'(-1)    s2  ==  c
  // T0'(1)  T1'(1)  T2'(1)  T3'(1)     s3      d

  Real s0 = 0.5*(a + b) + 0.125*(c - d);
  Real s1 = 0.5625*(b - a) - 0.0625*(c + d);
  Real s2 = 0.125*(d - c);
  Real s3 = 0.0625*(a - b + c + d);

  f[0] -= s0;
  f[1] -= s1;
  f[2] -= s2;
  f[3] -= s3;
  f.setBounds(ya,yb);
}

void fixDiri(ComplexChebyCoeff& f) {
  fixDiri(f.re);
  fixDiri(f.im);
}
void fixDiriMean(ComplexChebyCoeff& f) {
  fixDiriMean(f.re);
  fixDiriMean(f.im);
}

void fixDiriNeum(ComplexChebyCoeff& f) {
  fixDiriNeum(f.re);
  fixDiriNeum(f.im);
}


void field2vector(const FlowField& u, ColumnVector& a) {

  assert(u.xzstate() == Spectral && u.ystate() == Spectral);
  int Kx = u.kxmaxDealiased();
  int Kz = u.kzmaxDealiased();
  int Ny = u.Ny();

  int N = 4*(Kx+2*Kx*Kz+Kz)*(Ny-3)+2*(Ny-2);
  if (a.length() == 0)
    a.resize_and_fill(N,0);
  else
    setToZero(a);
  assert(a.length() >= N);


  int n=0;

  // These loops follow Gibson, Halcrow, Cvitanovic table 1.
  for (int ny=2; ny<Ny; ++ny)
    a(n++) = Re(u.cmplx(0,ny,0,0)); // Ny-2 modes (line 1 of table)

  for (int ny=2; ny<Ny; ++ny)
    a(n++) = Re(u.cmplx(0,ny,0,2)); // Ny-2 modes (line 2)

  // Some coefficients from the FlowField are linearly dependent due
  // to BCs and divergence constraint. Omit these from a(n), and reconstruct
  // them in vector2field(v,u) using the constraint equations. In what follows,
  // which coefficients are omitted is determined by the changing ny loop index
  // e.g. (ny=3; n<Ny-1; ++ny) omits ny=0,1 and Ny-1, and the constraint eqns
  // that allow their reconstruction are mentioned in comments to the right
  // of the loop. 2007-06-07 jfg.
  for (int kx=1; kx<=Kx; ++kx) {
    int mx = u.mx(kx);
    for (int ny=2; ny<Ny; ++ny) {      // w BCs => ny=0,1 coefficients
      a(n++) = Re(u.cmplx(mx,ny,0,2)); // J(Ny-2) modes (line 3a)
      a(n++) = Im(u.cmplx(mx,ny,0,2)); // J(Ny-2) modes (line 3b)
    }
    for (int ny=3; ny<Ny-1; ++ny) {    // u BCs => 0,1; v BC => 2; div => Ny-1
      a(n++) = Re(u.cmplx(mx,ny,0,0)); // J(Ny-4) modes (line 4a)
      a(n++) = Im(u.cmplx(mx,ny,0,0)); // J(Ny-4) modes (line 4b)
    }
  }
  for (int kz=1; kz<=Kz; ++kz) {
    int mz = u.mz(kz);
    for (int ny=2; ny<Ny; ++ny) {      // u BCs => 0,1; v BC => 2; div => Ny-1
      a(n++) = Re(u.cmplx(0,ny,mz,0)); // K(Ny-2)  (line 5a)
      a(n++) = Im(u.cmplx(0,ny,mz,0)); // K(Ny-2)  (line 5b)
    }
    for (int ny=3; ny<Ny-1; ++ny) {    // w BCs => 0,1; v BC => 2; div => Ny-1
      a(n++) = Re(u.cmplx(0,ny,mz,2)); // K(Ny-2)  (line 6a)
      a(n++) = Im(u.cmplx(0,ny,mz,2)); // K(Ny-2)  (line 6b)
    }
  }
  for (int kx=-Kx; kx<=Kx; ++kx) {
    if (kx==0)
      continue;

    int mx = u.mx(kx);
    for (int kz=1; kz<=Kz; ++kz) {
      int mz = u.mz(kz);
      for (int ny=2; ny<Ny; ++ny) {     // u BCs => 0,1;
	a(n++) = Re(u.cmplx(mx,ny,mz,0)); // JK(Ny-2)  (line 7a)
	a(n++) = Im(u.cmplx(mx,ny,mz,0)); // JK(Ny-2)  (line 7b)
      }
      for (int ny=3; ny<Ny-1; ++ny) { // w BCs => 0,1; v BC => 2; div => Ny-1
	a(n++) = Re(u.cmplx(mx,ny,mz,2)); // JK(Ny-4)  (line 8a)
	a(n++) = Im(u.cmplx(mx,ny,mz,2)); // JK(Ny-4)  (line 8b)
      }
    }
  }
}

void vector2field(const ColumnVector& a, FlowField& u) {

  assert(u.xzstate() == Spectral && u.ystate() == Spectral);
  int Kx=u.kxmaxDealiased();
  int Kz=u.kzmaxDealiased();
  int Ny=u.Ny(); // max polynomial order == number y gridpoints - 1
  Real ya = u.a();
  Real yb = u.b();
  Real Lx = u.Lx();
  Real Lz = u.Lz();

  assert(a.length() >= 4*(Kx+2*Kx*Kz+Kz)*(Ny-3)+2*(Ny-2));

  u.setToZero();

  Complex zero(0.0, 0.0);
  ComplexChebyCoeff f0(Ny, ya, yb, Spectral);
  ComplexChebyCoeff f1(Ny, ya, yb, Spectral);
  ComplexChebyCoeff f2(Ny, ya, yb, Spectral);

  // These loops follow Gibson, Halcrow, Cvitanovic table 1.
  // The vector contains the linearly independent real numbers in a
  // zero-div no-slip FlowField. The idea is to extract these from the
  // vector and reconstruct the others from the BCs and div constraint
  // See corresponding comments in field2vector(u,v);

  int n=0;

  // =========================================================
  // (0,0) Fourier mode
  for (int ny=2; ny<Ny; ++ny)
    f0.re[ny] = a(n++);
  fixDiri(f0.re);
  for (int ny=0; ny<Ny; ++ny)
    u.cmplx(0,ny,0,0) = f0[ny];

  for (int ny=2; ny<Ny; ++ny)
    f2.re[ny] = a(n++);
  fixDiri(f2.re);
  for (int ny=0; ny<Ny; ++ny)
    u.cmplx(0,ny,0,2) = f2[ny];

  // =========================================================
  // (kx,0) Fourier modes, 0<kx
  for (int kx=1; kx<=Kx; ++kx) {
    int mx = u.mx(kx);

    for (int ny=2; ny<Ny; ++ny) {
      f2.re[ny] = a(n++);
      f2.im[ny] = a(n++);
    }
    fixDiri(f2);

    for (int ny=3; ny<Ny-1; ++ny) {
      f0.re[ny] = a(n++);
      f0.im[ny] = a(n++);
    }
    f0.re[Ny-1] = 0.0;
    f0.im[Ny-1] = 0.0;
    fixDiriMean(f0);

    integrate(f0, f1);
    f1.sub(0, 0.5*(f1.eval_a() + f1.eval_b()));
    f1 *= Complex(0.0, -(2*pi*kx)/Lx);

    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mx,ny,0,0) = f0[ny];
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mx,ny,0,1) = f1[ny];
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mx,ny,0,2) = f2[ny];

    // ------------------------------------------------------
    // Now copy conjugates of u(kx,ny,0,i) to u(-kx,ny,0,i). These are
    // redundant modes stored only for the convenience of FFTW.
    int mxm = u.mx(-kx);
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mxm,ny,0,0) = conj(f0[ny]);
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mxm,ny,0,1) = conj(f1[ny]);
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mxm,ny,0,2) = conj(f2[ny]);

  }


  // =========================================================
  // (0,kz) Fourier modes, 0<kz.
  for (int kz=1; kz<=Kz; ++kz) {
    int mz = u.mz(kz);

    for (int ny=2; ny<Ny; ++ny) {
      f0.re[ny] = a(n++);
      f0.im[ny] = a(n++);
    }
    fixDiri(f0);
    for (int ny=3; ny<Ny-1; ++ny) {
      f2.re[ny] = a(n++);
      f2.im[ny] = a(n++);
    }
    f2.re[Ny-1] = 0.0;
    f2.im[Ny-1] = 0.0;
    fixDiriMean(f2);

    integrate(f2, f1);
    f1.sub(0, 0.5*(f1.eval_a() + f1.eval_b()));
    f1 *= Complex(0.0, -(2*pi*kz)/Lz);

    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(0,ny,mz,0) = f0[ny];
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(0,ny,mz,1) = f1[ny];
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(0,ny,mz,2) = f2[ny];
  }

  // =========================================================
  for (int kx=-Kx; kx<=Kx; ++kx) {
    if (kx==0)
      continue;

    int mx = u.mx(kx);
    for (int kz=1; kz<=Kz; ++kz) {
      int mz = u.mz(kz);

      for (int ny=2; ny<Ny; ++ny) {
	f0.re[ny] = a(n++);
	f0.im[ny] = a(n++);
      }
      fixDiri(f0);

      // f0 is complete. Load it into u.
      for (int ny=0; ny<Ny; ++ny)
	u.cmplx(mx,ny,mz,0) = f0[ny];

      for (int ny=3; ny<Ny-1; ++ny) {
	f2.re[ny] = a(n++);
	f2.im[ny] = a(n++);
      }
      f2.re[Ny-1] = -f0.re[Ny-1]*(kx*Lz)/(kz*Lx);
      f2.im[Ny-1] = -f0.im[Ny-1]*(kx*Lz)/(kz*Lx);

      // Adjust 0,1,2 coeffs of f2 so that f2(+/-1) = 0 and
      // kz/Lz mean(f2) + kx/Lx mean(f0) == 0 (so that  v(1)==v(-1))
      Complex f2a = f2.eval_a();
      Complex f2b = f2.eval_b();
      Complex f0m = f0.mean();
      Complex f2m = f2.mean() + (kx*Lz)/(kz*Lx)*f0m;

      f2.sub(0, 0.125*(f2a+f2b) + 0.75*f2m);
      f2.sub(1, 0.5*(f2b-f2a));
      f2.sub(2, 0.375*(f2a+f2b) - 0.75*f2m);

      for (int ny=0; ny<Ny; ++ny)
	u.cmplx(mx,ny,mz,2) = f2[ny];

      f0 *= Complex(0, -2*pi*kx/Lx);
      f2 *= Complex(0, -2*pi*kz/Lz);
      f0 += f2;

      integrate(f0, f1);
      f1.sub(0, 0.5*(f1.eval_a() + f1.eval_b()));

      for (int ny=0; ny<Ny; ++ny)
	u.cmplx(mx,ny,mz,1) = f1[ny];
    }
  }
}

void field2vector(const FlowField& u, ColumnVector& v, const std::vector<RealProfileNG>& basis) {
  if (basis.size() == 0)
    field2vector(u,v);
  else {
    if (v.length() == 0)
      v.resize_and_fill(basis.size(), 0);
    else
      setToZero(v);

    for (uint i=0; i<basis.size(); ++i)
      v(i) = L2InnerProduct(u, basis[i]);
  }
  return;
}


void vector2field(const ColumnVector& v, FlowField& u, const std::vector<RealProfileNG>& basis) {
  if (basis.size() == 0)
    vector2field(v,u);
  else {

    RealProfileNG prof;
    const Real EPSILON = 1e-16;
    u.setToZero();

    for (uint i=0; i<basis.size(); ++i) {
      if (abs(v(i)) < EPSILON)
	continue;
      prof = basis[i];
      prof *= v(i);
      u += prof;
    }
  }
}

void vector2field(const vector<RealProfile>& basis, const ColumnVector& a,
		  FlowField& u) {
  int N = basis.size();
  assert(a.length() >= N);
  u.setToZero();
  RealProfile tmp;

  const Real EPSILON = 1e-16;
  for (int n=0; n<N; ++n) {
    Real an = a(n);
    if (abs(an) < EPSILON)
      continue;

    tmp = basis[n];
    tmp *= an;

    assert(tmp.kx() <= u.kxmax() && tmp.kx() >= u.kxmin() &&
	   tmp.kz() <= u.kzmax() && tmp.kz() >= u.kzmin());

    u += tmp;
  }
}




void field2vector(const vector<RealProfile>& basis, const FlowField& u, ColumnVector& a) {

  int N = basis.size();
  if (a.length() == 0)
    a.resize_no_fill(N);
  assert(a.length() >= N);

  for (int n=0; n<N; ++n)
    a(n) = L2InnerProduct(u, basis[n]);
}



void vector2field(const vector<BasisFunc>& basis, const ComplexColumnVector& a,
		  FlowField& u) {
  int N = basis.size();
  assert(a.length() >= N);
  u.setToZero();
  BasisFunc tmp;

  const Real EPSILON = 1e-16;
  for (int n=0; n<N; ++n) {
    Complex an = a(n);
    if (abs(an) < EPSILON)
      continue;

    tmp = basis[n];
    tmp *= an;

    assert(tmp.kx() <= u.kxmax() && tmp.kx() >= u.kxmin() &&
	   tmp.kz() <= u.kzmax() && tmp.kz() >= u.kzmin());

    u += tmp;
  }
}

void quicksort(ComplexColumnVector& z) {
  int N = z.length();
  array<Real> x(N);
  for (int n=0; n<N; ++n)
    x[n] = abs(z(n));
  array<int> i = quicksort(x);
  ComplexColumnVector w(N);
  for (int n=0; n<N; ++n)
    w(n) = z(i[N-n-1]);
  z = w;
}

// Does extra copying
void quicksort(ComplexColumnVector& Lambda, ComplexMatrix& V) {
  int M = V.rows();
  int N = V.cols();
  assert(Lambda.length() == N);

  array<Real> magn(N);
  for (int n=0; n<N; ++n)
    magn[n] = abs(Lambda(n));
  array<int> i = quicksort(magn);

  ComplexColumnVector Lambda_sorted(N);
  ComplexMatrix V_sorted(M,N);

  for (int n=0; n<N; ++n) {
    Lambda_sorted(n) = Lambda(i[N-n-1]);
    V_sorted.insert(V.column(i[N-n-1]),0,n);
  }
  Lambda = Lambda_sorted;
  V = V_sorted;
}

void leastsquares(const Matrix& Hn, ColumnVector& bk, ColumnVector& y,
		  Real& residual) {
  //cout << "entering leastsquares..." << endl;

  // Hn y = bk, Hn = QR
  // QQR y = bk
  //  R y = Q' bk;

  QR qr(Hn, QR::economy);
  Matrix Q = qr.Q();
  Matrix R = qr.R();
  int N = R.rows();
  RowVector c = bk.transpose()*Q;

  // Backsolve R y = Q* b
  for (int i=N-1; i>=0; --i) {
    Real sum = c(i);
    for (int j=i+1; j<N; ++j)
      sum -= R(i,j)*y(j);
    y(i) = sum/R(i,i);
  }

  residual = L2Norm(Hn*y-bk)/L2Norm(bk);
  //cout << "subspace residual = " << L2Norm(R*y-c)/L2Norm(c) << endl;
  //cout << "leastsqs residual = " << residual << endl;
  //cout << "L2Norm(bk)        == " << L2Norm(bk) << endl;
  //cout << "L2Norm(Hn*y-bk)   == " << L2Norm(Hn*y-bk) << endl;
  //cout << "L2Norm(Q'*bk)     == " << L2Norm(c) << endl;
  //cout << "L2Norm(R*y-Q'*bk) == " << L2Norm(R*y-c) << endl;
}

Arnoldi::Arnoldi()
  :
  M_(0),
  Niter_(0),
  n_(0),
  condition_(0),
  H_(0,0),
  Q_(0,0),
  Vn_(0,0),
  qn_(0),
  v_(0),
  ew_(0),
  ev_(0,0)
{}

Arnoldi::Arnoldi(const ColumnVector& b, int Niterations, Real minCondition)
  :
  M_(b.length()),
  Niter_(Niterations),
  n_(0),
  condition_(minCondition),
  H_(Niter_+1, Niter_),
  Q_(M_, Niter_+1),
  Vn_(0,0),
  qn_(b),
  v_(M_),
  ew_(0),
  ev_(0,0)
{
  setToZero(H_);
  setToZero(Q_);
  qn_ *= 1.0/L2Norm(qn_);
  qn_ *= 1.0/L2Norm(qn_);
  Q_.insert(qn_,0,0);
}

void Arnoldi::iterate(const ColumnVector& Aq) {
  if (n_ == Niter_) {
    cerr << "warning : Arnoldi::iterate(Aq) : \n"
	 << "reached maximum number of iterations. doing nothing.\n";
  }
  v_ = Aq;

  // Orthogonalize v and insert it in Q matrix
  //cout << "calculating arnoldi iterate..." << endl;
  for (int j=0; j<=n_; ++j) {
    //cout << j << ' ' << flush;
    ColumnVector Qj = Q_.column(j);
    H_(j,n_) = Qj.transpose() * v_;
    v_ -= H_(j,n_)*Qj;
  }
  //cout << endl;
  Real vnorm = L2Norm(v_);

  if (abs(vnorm) < condition_) {
    H_(n_+1, n_) = 0.0;
    v_ *= 0.0;
    cerr << "Arnoldi breakdown. Exiting\n";
    exit(1);
  }

  else {
    H_(n_+1, n_) = vnorm;
    v_ *= 1.0/vnorm;
  }
  Q_.insert(v_,0,n_+1);
  qn_ = Q_.column(n_+1);
  ++n_;
}

const ColumnVector& Arnoldi::testVector() const {
  return qn_;
}

int Arnoldi::n() const {return n_;}
int Arnoldi::Niter() const {return Niter_;}

const ComplexColumnVector& Arnoldi::ew() {
  if (ew_.length() != n_)
    eigencalc();
  return ew_;
}

void Arnoldi::eigencalc() {
  Matrix Qn  = Q_.extract(0,0,M_-1,n_-1);
  Matrix Hn = H_.extract(0,0,n_-1,n_-1);
  //cout << "Q == \n" << Q_ << endl;
  //cout << "H == \n" << H_ << endl;
  //cout << "Hn == \n" << Hn << endl;
  //  eignHn_ = EIG(Hn);
  EIG eignHn(Hn);
  ew_ = eignHn.eigenvalues();
  ComplexMatrix W = eignHn.eigenvectors();
  quicksort(ew_, W);
  ev_ = Qn*W;
};

const ComplexMatrix& Arnoldi::ev() {
  if (ev_.length() != n_)
    eigencalc();
  return ev_;
}


GMRES::GMRES()
  :
  M_(0),
  Niter_(0),
  n_(0),
  condition_(0),
  H_(0,0),
  Q_(0,0),
  Vn_(0,0),
  x0_(0),
  v_(0),
  qn_(0),
  bnorm_(0)
{}

GMRES::GMRES(const ColumnVector& b, int Niterations, Real minCondition)
  :
  M_(b.length()),
  Niter_(Niterations),
  n_(0),
  condition_(minCondition),
  H_(Niter_+1, Niter_),
  Q_(M_, Niter_+1),
  Vn_(0,0),
  x0_(0),
  v_(0),
  qn_(b),
  bnorm_(L2Norm(b))
{
  setToZero(H_);
  setToZero(Q_);
  qn_ *= 1.0/L2Norm(qn_);
  qn_ *= 1.0/L2Norm(qn_);
  Q_.insert(qn_,0,0);
}

/**********************************
GMRES::GMRES(const ColumnVector& guess, const ColumnVector& b, int Niterations, Real minCondition)
  :
  M_(b.length()),
  Niter_(Niterations),
  n_(0),
  condition_(minCondition),
  H_(Niter_+1, Niter_),
  Q_(M_, Niter_+1),
  Vn_(0,0),
  x0_(guess),
  v_(0),
  qn_(b),
  bnorm_(L2Norm(b))
{
  if (guess.length() != b.length()) {
    cerr << "error in : GMRES::GMRES(guess,b,Niterations,minCondition)" << endl;
    cerr << "guess.length() != b.length()). Exiting." << endl;
  }
  setToZero(H_);
  setToZero(Q_);
  qn_ *= 1.0/L2Norm(qn_);
  qn_ *= 1.0/L2Norm(qn_);
  Q_.insert(qn_,0,0);
}
************************************/

void GMRES::iterate(const ColumnVector& Aq) {
  if (n_ == Niter_) {
    cerr << "warning : GMRES::iterate(Ab) : \n"
	 << "reached maximum number of iterations. doing nothing\n";
    return;
  }
  v_ = Aq;

  // Orthogonalize v and insert it in Q matrix
  //cout << "calculating arnoldi iterate..." << endl;
  for (int j=0; j<=n_; ++j) {
    //cout << j << ' ' << flush;
    ColumnVector Qj = Q_.column(j);
    H_(j,n_) = Qj.transpose() * v_;
    v_ -= H_(j,n_)*Qj;
  }
  //cout << endl;
  Real vnorm = L2Norm(v_);

  // If v is within subspace spanned by Q, replace it with a random vector.
  // There's a measure 0 chance that any particular random vector will also
  // spanned by Q, so allow for a limited number of retries.
  int maxretries = 10;
  int retries = 0;
  while (abs(vnorm) < condition_ && retries++ < maxretries) {
    cerr << "GMRES benign breakdown: \n"
	 << "  Using random vector instead of linearly dependent Krylov vector\n.";

    for (int i=0; i<v_.length(); ++i)
      v_(i) = randomReal();

    // Gram-Schmidt orthogonalize v w.r.t Q
    for (int j=0; j<=n_; ++j) {
      ColumnVector Qj = Q_.column(j);
      H_(j,n_) = Qj.transpose() * v_;
      v_ -= H_(j,n_)*Qj;
    }
    //cout << endl;
    vnorm = L2Norm(v_);
  }

  H_(n_+1, n_) = vnorm;
  v_ *= 1.0/vnorm;
  Q_.insert(v_,0,n_+1);
  qn_ = Q_.column(n_+1);

  Matrix Hn = H_.extract(0,0,n_+1,n_);
  Matrix Qn = Q_.extract(0,0,M_-1,n_);
  ColumnVector bk(n_+2);
  setToZero(bk);
  bk(0) = bnorm_;

  ColumnVector y(n_+1);
  leastsquares(Hn,bk,y,residual_);
  xn_ = Qn*y;
  ++n_;
}

int GMRES::n() const {return n_;}
int GMRES::Niter() const {return Niter_;}

const ColumnVector& GMRES::testVector() const {
  return qn_;
}
const ColumnVector& GMRES::solution() const{
  return xn_;
}
Real GMRES::residual() const{
  return residual_;
}
Matrix GMRES::Hn() const {
  return H_.extract(0,0,n_,n_-1);
}
Matrix GMRES::Qn() const {
  return Q_.extract(0,0,M_-1,n_-1);
}
Matrix GMRES::Qn1() const {
  return Q_.extract(0,0,M_-1,n_);
}

/******************************************
HookStepSearch::HookStepSearch()
  :
  newtonStep_(0),
  stuck_(false),
  residual_(0),
  delta_(0),
  x_(0),
  fx_(0),
{}

HookStepSearch::HookStepSearch(Rn2Rnfunc, const ColumnVector& x0);
  :
  newtonStep_(0),
  residual_(0),
  delta_(0),
  stuck_(false),
  Nunk_(0),
  x_(0),
  fx_(0),
{}


HookStepSearch::HookStepSearch(Rn2Rnfunc, const ColumnVector& x0);const HookstepParams& p) : params_(p) {
  newtonStep_(0),
  stuck_(false),
  residual_(0),
  Nunk_(0),
  delta_(0),
  x_(0),
  fx_(0),
}
****************************************************/

ColumnVector hookstepSearch(Rn2Rnfunc f, const ColumnVector& x0) {
  HookstepParams p;
  return hookstepSearch(f, x0, p);
}

HookstepParams::HookstepParams(Real epsSearch_,
			       Real epsDx_,
			       bool centdiff_,
			       int  Nnewton_,
			       int  Nhook_,
			       Real deltaInit_,
			       Real deltaMin_,
			       Real deltaMax_,
			       Real deltaFuzz_,
			       Real lambdaMin_,
			       Real lambdaMax_,
			       Real lambdaRequiredReduction_,
			       Real improvReq_,
			       Real improvOk_,
			       Real improvGood_,
			       Real improvAcc_,
			       bool pausing_)
  :
  epsSearch(epsSearch_),
  epsDx(epsDx_),
  centdiff(centdiff_),
  Nnewton(Nnewton_),
  Nhook(Nhook_),
  deltaInit(deltaInit_),
  deltaMin(deltaMin_),
  deltaMax(deltaMax_),
  deltaFuzz(deltaFuzz_),
  lambdaMin(lambdaMin_),
  lambdaMax(lambdaMax_),
  lambdaRequiredReduction(lambdaRequiredReduction_),
  improvReq(improvReq_),
  improvOk(improvOk_),
  improvGood(improvGood_),
  improvAcc(improvAcc_),
  xscale(0),
  fscale(0),
  pausing(pausing_)
{}

ColumnVector hookstepSearch(Rn2Rnfunc f, const ColumnVector& x0, const HookstepParams& p) {

  Real epsSearch = p.epsSearch; // stop search if |f(x)| < epsSearch
  Real epsDx     = p.epsDx;     // scale for perturbs in finite-diff estimate of Df
  bool centdiff  = p.centdiff;// used centered differences in estimating Df
  int  Nnewton   = p.Nnewton;   // max number of Newton step
  int  Nhook     = p.Nhook;     // max number of hookstep iterations per Newton
  Real deltaInit = p.deltaInit; // initial trust region radius
  Real deltaMin  = p.deltaMin;  // stop if radius of trust region gets this small
  Real deltaMax  = p.deltaMax;  // don't let trust region radius get bigger than this
  Real deltaFuzz = p.deltaFuzz; // accept steps within (1+/-deltaFuzz)*delta
  Real lambdaMin = p.lambdaMin; // minimum delta shrink rate
  Real lambdaMax = p.lambdaMax; // maximum delta expansion rate
  Real lambdaRequiredReduction = p.lambdaRequiredReduction;
  Real improvReq = p.improvReq;
  Real improvOk   = p.improvOk;
  Real improvGood = p.improvGood;
  Real improvAcc  = p.improvAcc;
  ColumnVector xscale = p.xscale;
  ColumnVector fscale = p.fscale;
  bool pausing = p.pausing;

  string outdir = "./";

  Real delta = deltaInit;
  int fcount_newton = 0;
  int fcount_hookstep = 0;

  const int Nunk = x0.length();
  const int Neqn = Nunk;
  int label;
  bool cacheable;

  ofstream xnewtos("xnewt.asc");
  ofstream dnewtos("dxnewt.asc");
  ofstream fnewtos("fnewt.asc");
  ofstream osconv("convergence.asc");
  osconv.setf(ios::left);
  osconv << setw(14) << "% residual" << setw(14) << "delta"
	 << setw(14) << "L2Norm(dx)" << setw(14) << "L2Norm(sN)"
	 << setw(14) << "L2Norm(sH)" << setw(10) << "ftotal"
	 << setw(10) << "fnewt" << setw(10) << "fhook" << endl;

  // ==============================================================
  // Initialize Newton iteration
  cout << "Computing f(x)" << endl;
  ColumnVector       x(x0);
  ColumnVector      fx = (*f)(x, label=0, cacheable=true); // 0 ==> current best guess

  // From here on, scale/unscale x when interfacing with the external world
  rescale(x,xscale);

  ColumnVector      dx(Nunk); // temporary for increments in unknowns
  ColumnVector   Df_dx(Nunk); // "Df*dx" = 1/eps (f(x + eps dx)
  ColumnVector      xS(Nunk); // stores xN and xH, the Newton and Hookstep unknowns
  ColumnVector      fS(Nunk); // stores fN and fH, fS = f(xS);
  ColumnVector      sN(Nunk); // the Newton step

  setToZero(dx);
  //for (i=0; i<Nunk; ++i)
  //dx(i) = 0.0;

  Real rx = 0.0;        // Dennis & Schnabel residual r(x) = 1/2 ||f(x)||^2
  Real init_rx = 0.0;
  Real prev_rx = 0.0;
  Real snorm   = 0.0;   // norm of hookstep
  Real sNnorm  = 0.0;   // norm of newton step
  Real dxnorm  = 0.0;
  Real xnorm   = 0.0;
  bool stuck = false;

  for (int newtonStep=0; newtonStep<Nnewton; ++ newtonStep) {

    cout << "========================================================" << endl;
    cout << "Newton iteration number " << newtonStep << endl;

    // Compute quantities that will be used multiple times
    // fx     = f(x) - x
    dxnorm = L2Norm(dx);
    xnorm = L2Norm(x);
    rx = 0.5*L2Norm2(fx);

    cout << "Current state of Newton iteration:" << endl;
    cout << "fcount_newton  == " << fcount_newton << endl;
    cout << "fcount_hookstep== " << fcount_hookstep << endl;
    cout << "    L2Norm(x)  == " << xnorm << endl;
    cout << "   L2Norm(dx)  == " << dxnorm <<endl;
    cout << "   L2Norm(sH)  == " << snorm <<endl;
    //cout << "L2Dist(x,x0)   == " << L2Dist(x, x0) << endl;
    cout << "rx == 1/2 Norm2(f(x)) : " << endl;
    cout << "   initial  rx == " << init_rx << endl;
    cout << "   previous rx == " << prev_rx << endl;
    cout << "   current  rx == " << rx << endl;
    cout << "             x == ";
    for (int i=0; i<Nunk; ++i) cout << x(i) << ' ';  cout << endl;
    cout << "            dx == ";
    for (int i=0; i<Nunk; ++i) cout << dx(i) << ' '; cout << endl;
    cout << "            fx == ";
    for (int i=0; i<Nunk; ++i) cout << fx(i) << ' '; cout << endl;
    cout << "        delta  == " << delta << endl;

    if (newtonStep == 0) {
      init_rx = rx;
      prev_rx = rx;
    }

    osconv.setf(ios::left);
    osconv.setf(ios::left); osconv << setw(14) << rx;
    osconv.setf(ios::left); osconv << setw(14) << delta;
    osconv.setf(ios::left); osconv << setw(14) << dxnorm;
    osconv.setf(ios::left); osconv << setw(14) << sNnorm;
    osconv.setf(ios::left); osconv << setw(14) << snorm;
    osconv.setf(ios::left); osconv << setw(10) << fcount_newton + fcount_hookstep;
    osconv.setf(ios::left); osconv << setw(10) << fcount_newton;
    osconv.setf(ios::left); osconv << setw(10) << fcount_hookstep;
    osconv << endl;

    unscale( x,xscale);
    unscale(dx,xscale);
    save(x, outdir + "xnewt" + i2s(newtonStep));
    save(x, outdir + "xbest");
    for (int i=0; i<Nunk; ++i) {
      xnewtos << x(i) << ' ';
      dnewtos << dx(i) << ' ';
      fnewtos << fx(i) << ' ';
    }
    xnewtos << endl;
    dnewtos << endl;
    fnewtos << endl;
    rescale( x,xscale);
    rescale(dx,xscale);

    if (rx < epsSearch) {
      cout << "Newton search converged. Breaking." << endl;
      cout << "1/2 L2Norm(f(x)) == " << rx << " < "
	   << epsSearch << " == " << epsSearch << endl;
      break;
    }
    else if (stuck) {
      cout << "Newton search is stuck. Breaking." << endl;
      break;
    }
    else if (newtonStep == Nnewton) {
      cout << "Reached maximum number of Newton steps. Breaking." << endl;
      break;
    }
    if (pausing)
      cfpause();

    prev_rx = rx;

    // Set up RHS vector b = -f(x)
    ColumnVector b(Neqn);
    setToZero(b);
    b = fx; // field2vector(fu, b, basis);
    b *= -1.0;

    // Estimate f_ij = df_i/dx_j with finite differences
    //               = 1/eps (f_i(x + eps e_j) - f_i(x))
    Matrix Df(Nunk, Nunk);

    for (int j=0; j<Nunk; ++j) {
      ColumnVector f_x_dxj(Nunk);
      ColumnVector x_dxj = x;
      x_dxj(j) += epsDx;
      unscale(x_dxj, xscale);
      f_x_dxj = (*f)(x_dxj, label=-1, cacheable=false); // -1 ==> evaluating Df

      for (int i=0; i<Nunk; ++i)
	Df(i,j) = (f_x_dxj(i)- fx(i))/epsDx;
    }
    fcount_newton += Nunk*Nunk;

    save(Df, "Df");

    SVD svd(Df, SVD::std);

    Matrix V = svd.right_singular_matrix();
    Matrix Ut = svd.left_singular_matrix().transpose();
    DiagMatrix D = svd.singular_values();
    cout << "D == ";
    for (int i=0; i<D.rows(); ++i)
      cout << D(i,i) << ' ';
    cout << endl;
    //cfpause();

    solve(Ut, D, V, b, sN);
    sNnorm = L2Norm(sN);

    cout << "newton step sN == " << sN(0) << ' ' << sN(1) << endl;

    // ==================================================================
    // Hookstep algorithm

    cout << "------------------------------------------------" << endl;
    cout << "Beginning hookstep calculations." << endl;

    int Nk = Nunk;

    ColumnVector bh = Ut*b; // b hat

    int hookcount = 0;
    bool have_backup  = false;
    Real backup_snorm = 0.0;
    Real backup_delta = 0.0;
    Real backup_rS    = 0.0;

    ColumnVector backup_dx;
    ColumnVector backup_fS;

    ColumnVector sH;

    //bool deltahas_decreased = false;
    Real deltaMaxLocal = deltaMax;

    // Find a good trust region and the optimal hookstep in it.
    // Quality of trust region
    while (true) {

      cout << "-------------------------------------------" << endl;
      cout << "Newton, hookstep number " << newtonStep << ", "
	   << hookcount++ << endl;
      cout << "delta == " << delta << endl;

      bool hookstep_equals_newtonstep;

      if (L2Norm(sN) <= delta) {
	cout << "Newton step is within trust region: " << endl;
	cout << "L2Norm(sN) == " << L2Norm(sN) << " <= " << delta << " == delta" << endl;
	hookstep_equals_newtonstep = true;
	sH = sN;
      }
      else {
	cout << "Newton step is outside trust region: " << endl;
	cout << "L2Norm(sN) == " << L2Norm(sN) << " > " << delta << " == delta" << endl;
	cout << "Calculate hookstep sH with radius L2Norm(sH) == delta" << endl;

	hookstep_equals_newtonstep = false;

	// This for-loop determines the hookstep. Search for value of mu that
	// produces ||shmu|| == delta. That provides optimal reduction of
	// quadratic model of residual in trust region norm(shmu) <= delta.
	Real mu = 0.0;
	Real nshmu = 0;
	ColumnVector shmu(Nk); // (s hat)(mu)]

	for (int hookSearch=0; hookSearch<Nhook; ++hookSearch) {
	  for (int i=0; i<Nk; ++i)
	    shmu(i) = bh(i)/(D(i,i) + mu); // research notes
	  nshmu = L2Norm(shmu);
	  Real Phi = nshmu*nshmu - delta*delta;
	  cout << "mu, L2Norm(sH(mu)) == " << mu << ", " << nshmu << endl;

	  // Found satisfactory value of mu and thus shmu and sH s.t. |sH| < delta
	  if (nshmu < delta || (nshmu > (1-deltaFuzz)*delta && nshmu < (1+deltaFuzz)*delta))
	    break;

	  // Update value of mu and try again. Update rule is a Newton search for
	  // Phi(mu)==0 based on a model of form a/(b+mu) for norm(sh(mu)). See
	  // Dennis & Schnabel.
	  else if (hookSearch<Nhook-1) {
	    Real PhiPrime = 0.0;
	    for (int i=0; i<Nk; ++i) {
	      Real di_mu = D(i,i) + mu;
	      Real bi = bh(i);
	      PhiPrime -= 2*bi*bi/(di_mu*di_mu*di_mu);
	    }
	    mu -= (nshmu/delta) * (Phi/PhiPrime);
	  }
	  else {
	    cout << "Couldn't find solution of hookstep optimization eqn Phi(mu)==0" << endl;
	    cout << "This shouldn't happen. It indicates an error in the algorithm." << endl;
	    cout << "Exiting.\n";
	    exit(1);
	  }
	} // search over mu for norm(s) == delta

	cout << "Found hookstep of proper radius" << endl;
	cout << "nshmu == " << L2Norm(shmu) << " ~= " << delta << " == delta " << endl;
	sH = V*shmu;
      } // else {Newton step outside trust region, so

      // Compute
      // (1) actual residual from evaluation of
      //     rS == r(s+ds) act  == 1/2 (f(x+dx),f(x+dx))
      //
      // (2) predicted residual from quadratic model of r(s)
      //     rP == r(s+ds) pred  == 1/2 (f(x),f(x)) + (f(x), Df dx)
      //
      // (3) slope of r(s)
      //     dr/ds == (r(s + ds) - r(s))/|ds|
      //           == (f(x,T), Df dx /|ds|
      // where ( , ) is Vector inner product V2IP, which equals L2 norm of vector rep

      // (1) Compute actual residual of step, rS
      cout << "Computing residual of hookstep sH" << endl;
      dx = sH;
      dxnorm = L2Norm(dx);
      xS  = x;
      xS += dx;

      cout << "L2Norm(dx) == " << L2Norm(dx) << endl;
      unscale(xS, xscale);
      fS = (*f)(xS, newtonStep+1, cacheable=true);
      rescale(xS, xscale);

      cout << endl;

      Real rS = 0.5*L2Norm2(fS);       // actual residual of hookstep
      Real Delta_rS = rS - rx;         // improvement in residual
      //cout << "r(s), r(s+ds) == " << rx << ", " << rS << endl;

      // (2) and (3) Compute quadratic model and slope
      cout << "Computing local quadratic model of residual" << endl;
      ColumnVector Df_dx = Df*dx;
      cout << endl;
      // Local quadratic and linear models of residual, based on Taylor exp at current position
      // Quadratic model of r(s+ds): rQ == 1/2 |f(x) + Df dx|^2
      // Linear  model   of r(s+ds): rL == 1/2 (f(x),f(x)) + (f(x), Df dx)

      Real Delta_rL = L2IP(fx,Df_dx); // rL  - 1/2 (f(x),f(x)) == (f(x), Df dx)
      Real rL = rx + Delta_rL;        // rL == 1/2 (f(x),f(x))  + (f(x), Df dx)

      if (rL >= rx) {
	cout << "error : local linear model of residual is increasing, indicating\n"
	     << "        that the solution to the Newton equations is inaccurate\n"
	     << "exiting." << endl;
	exit(1);
      }

      Df_dx += fx;
      Real rQ = 0.5*L2Norm2(Df_dx);  // rQ == 1/2 |f(x) + Df dx|^2
      Real Delta_rQ = rQ - rx;       // rQ == 1/2 |(f(x)|^2 - 1/2 |f(x) + Df dx|^2

      snorm = L2Norm(sH);

      // ===========================================================================
      cout << "Determining what to do with current Newton/hookstep and trust region" << endl;

      // Coefficients for non-local quadratic model of residual, based on r(s), r'(s) and r(d+ds)
      // dr/ds == (f(x), Df dx)/|ds|
      Real drds  = Delta_rL/snorm;

      // Try to minimize a quadratic model of the residual
      // r(s+ds) = r(s) + r' |ds| + 1/2 r'' |ds|^2
      Real lambda = -0.5*drds*snorm/(rS - rx - drds*snorm);

      // Compare the actual reduction in residual to the quadratic model of residual.
      // How well the model matches the actual reduction will determine, later on, how and
      // when the radius of the trust region should be changed.
      Real Delta_rS_req  = improvReq*drds*snorm;    // the minimum acceptable change in residual
      Real Delta_rS_ok   = improvOk*Delta_rQ;       // acceptable, but reduce trust region for next newton step
      Real Delta_rS_good = improvGood*Delta_rQ;     // acceptable, keep same trust region in next newton step
      Real Delta_rQ_acc  = abs(improvAcc*Delta_rQ); // for accurate models, increase trust region and recompute hookstep
      Real Delta_rS_accP = Delta_rQ + Delta_rQ_acc; //   upper bound for accurate predictions
      Real Delta_rS_accM = Delta_rQ - Delta_rQ_acc; //   lower bound for accurate predictions

      // Characterise change in residual Delta_rS
      ResidualImprovement improvement;  // fine-grained characterization

      // Place improvement in contiguous spectrum: Unacceptable > Poor > Ok > Good.
      if (Delta_rS > Delta_rS_req)
	improvement = Unacceptable;     // not even a tiny fraction of linear rediction
      else if (Delta_rS > Delta_rS_ok)
	improvement = Poor;             // worse than small fraction of quadratic prediction
      else if (Delta_rS < Delta_rS_ok && Delta_rS > Delta_rS_good)
	improvement = Ok;
      else {
	improvement = Good;             // not much worse or better than large fraction of prediction
	if (Delta_rS_accM <= Delta_rS &&  Delta_rS <= Delta_rS_accP)
	  improvement = Accurate;         // close to quadratic prediction
	else if (Delta_rS < Delta_rL)
	  improvement = NegaCurve;        // negative curvature in r(|s|) => try bigger step
      }
      int w = 14;

      cout << " rx      == " << setw(w) << rx            << " residual at current position" << endl;
      cout << " rS      == " << setw(w) << rS            << " residual of newton/hookstep" << endl;
      cout << "Delta_rS == " << setw(w) << Delta_rS      << " actual improvement in residual from newton/hookstep" << endl;
      cout << endl;

      if (improvement==Unacceptable)
      cout << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Unacceptable <------ " << endl;
      cout << "            " << setw(w) << Delta_rS_req  << " lower bound for acceptable improvement" << endl;
      if (improvement==Poor)
      cout << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Poor <------" << endl;
      cout << "            " << setw(w) << Delta_rS_ok   << " upper bound for ok improvement." << endl;
      if (improvement==Ok)
      cout << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Ok <------" << endl;
      cout << "            " << setw(w) << Delta_rS_good << " upper bound for good improvement." << endl;
      if (improvement==Good)
      cout << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Good <------" << endl;
      cout << "            " << setw(w) << Delta_rS_accP << " upper bound for accurate prediction." << endl;
      if (improvement==Accurate)
      cout << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Accurate <------" << endl;
      cout << "            " << setw(w) << Delta_rS_accM << " lower bound for accurate prediction." << endl;
      cout << "            " << setw(w) << Delta_rL      << " local linear model of improvement" << endl;
      if (improvement==NegaCurve)
      cout << "Delta_rS == " << setw(w) << Delta_rS      << " ------> NegativeCurvature <------" << endl;

      bool recompute_hookstep = false;

      cout << "lambda       == " << lambda << " is the reduction/increase factor for delta suggested by quadratic model" << endl;
      cout << "lambda*delta == " << lambda*delta << " is the delta suggested by quadratic model" << endl;

      // See comments at end for outline of this control structure

      // Following Dennis and Schnable, if the Newton step lies within the trust region,
      // reset trust region radius to the length of the Newton step, and then adjust it
      // further according to the quality of the residual.
      //if (hookstep_equals_newtonstep)
      //delta = snorm;

      // CASE 1: UNACCEPTABLE IMPROVEMENT
      // Improvement is so bad (relative to gradient at current position) that we'll
      // in all likelihood get better results in a smaller trust region.
      if (improvement == Unacceptable) {
	cout << "Improvement is unacceptable." << endl;

	if (have_backup) {
	  cout << "But we have a backup step that was acceptable." << endl;
	  cout << "Revert to backup step and backup delta, take the step, and go to next Newton iteration." << endl;
	  dx = backup_dx;
	  fS = backup_fS;
	  rS = backup_rS;
	  snorm = backup_snorm;
	  delta = backup_delta;

	  recompute_hookstep = false;
	}
	else {
	  cout << "No backup step is available." << endl;
	  cout << "Reduce trust region by minimizing local quadratic model and recompute hookstep." << endl;
	  deltaMaxLocal = delta;
	  lambda = adjustLambda(lambda, lambdaMin, lambdaRequiredReduction);
	  delta  = adjustDelta(delta, lambda, deltaMin, deltaMax);

	  if (delta > snorm) {
	    cout << "That delta is still bigger than the Newton step." << endl;
	    //cout << "This shouldn't happen. Control structure needs review." << endl;
	    cout << "Reducing delta to half the length of the Newton step." << endl;
	    cout << "  old delta == " << delta << endl;
	    delta = 0.5*snorm;
	    cout << "  new delta == " << delta << endl;
	  }
	  //delta_has_decreased = true;
	  recompute_hookstep = true;
	}
      }

      // CASE 2: EXCELLENT IMPROVEMENT AND ROOM TO GROW
      // Improvement == Accurate or Negacurve means we're likely to get a significant improvement
      // by increasing trust region. Increase trust region by a fixed factor (rather than quadratic
      // model) so that trust-region search is monotonic.
      // Exceptions are
      //   have_backup && backup_rS < rS -- residual is getting wrose as we increase delta
      //   hookstep==newtonstep          -- increasing trust region won't change answer
      //   delta>=deltamax               -- we're already at the largest allowable trust region
      else if ((improvement == NegaCurve || improvement == Accurate)
	       && !(have_backup && backup_rS < rS)
	       && !hookstep_equals_newtonstep
	       && !(delta >= deltaMax)) {

	cout << "Improvement is " << improvement << endl;

	//lambda = adjustLambda(lambda, lambdaMin, lambdaMax);
	Real new_delta = adjustDelta(delta, lambdaMax, deltaMin, deltaMax);

	if (new_delta < deltaMaxLocal) {
	  cout << "Continue adjusting trust region radius delta because" << endl;
	  cout << "Suggested delta has room: new_delta < deltaMaxLocal " << new_delta << " < " << deltaMaxLocal << endl;
	  if (have_backup)
	    cout << "And residual is improving:     rS < backup_sS     " << rS << " < " << backup_rS << endl;

	  cout << "Increase delta and recompute hookstep." << endl;
	  have_backup = true;
	  backup_dx = dx;
	  backup_fS = fS;
	  backup_rS = rS;
	  backup_snorm = snorm;
	  backup_delta = delta;

	  recompute_hookstep = true;
	  cout << " old delta == " << delta << endl;
	  delta = new_delta;
	  cout << " new delta == " << delta << endl;
	}
	else {
	  cout << "Stop adjusting trust region radius delta and take step because the new delta" << endl;
	  cout << "reached a local limit:  new_delta >= deltaMaxLocal " << new_delta << " >= " << deltaMaxLocal << endl;
	  cout << "Reset delta to local limit and go to next Newton iteration" << endl;
	  delta = deltaMaxLocal;
	  cout << "  delta == " << delta << endl;
	  recompute_hookstep = false;
	}
      }

      // CASE 3: MODERATE IMPROVEMENT, NO ROOM TO GROW, OR BACKUP IS BETTER
      // Remaining cases: Improvement is acceptable: either Poor, Ok, Good, or
      // {Accurate/NegaCurve and (backup is superior || Hookstep==NewtonStep || delta>=deltaMaxLocal)}.
      // In all these cases take the current step or backup if better.
      // Adjust delta according to accuracy of quadratic prediction of residual (Poor, Ok, etc).
      else {
	cout << "Improvement is " << improvement <<  " (some form of acceptable)." << endl;
	cout << "Stop adjusting trust region and take a step because" << endl;
	cout << "Improvement is merely Poor, Ok, or Good : " << (improvement == Poor || improvement == Ok || improvement == Good) << endl;
	cout << "Newton step is within trust region      : " << hookstep_equals_newtonstep << endl;
	cout << "Backup step is better                   : " << (have_backup && backup_rS < rS) << endl;
	cout << "Delta has reached local limit           : " << (delta >= deltaMax) << endl;

	recompute_hookstep = false;

	// Backup step is better. Take it instead of current step.
	if (have_backup && backup_rS < rS) {
	  cout << "Take backup step and set delta to backup value." << endl;
	  dx = backup_dx;
	  fS = backup_fS;
	  rS = backup_rS;

	  cout << "  old delta == " << delta << endl;
	  snorm = backup_snorm;
	  delta = backup_delta;
	  cout << "  new delta == " << delta << endl;
	}

	// Current step is best available. Take it. Adjust delta according to Poor, Ok, etc.
	// Will start new monotonic trust-region search in Newton-hookstep, so be more
	// flexible about adjusting delta and use quadratic model.
	else {
	  if (have_backup) {
	    cout << "Take current step and keep current delta, since it's produced best result." << endl;
	    cout << "delta == " << delta << endl;
	    // Keep current delta because current step was arrived at through a sequence of
	    // adjustments in delta, and this value yielded best results.
	  }
	  else if (improvement == Poor) {
	    cout << "Take current step and reduce delta, because improvement is poor." << endl;
	    lambda = adjustLambda(lambda, lambdaMin, 1.0);
	    delta  = adjustDelta(delta, lambda, deltaMin, deltaMax);
	  }
	  else if (improvement == Ok || hookstep_equals_newtonstep || (delta >= deltaMax)) {
	    cout << "Take current step and leave delta unchanged, for the following reasons:" << endl;
	    cout << "improvement            : " << improvement << endl;
	    cout << "|newtonstep| <= delta  : " << (hookstep_equals_newtonstep ? "true" : "false") << endl;
	    cout << "delta >= deltaMax      : " << (delta >= deltaMax    ? "true" : "false") << endl;
	    //cout << "delta has decreased    : " << (delta_has_decreased ? "true" : "false") << endl;
	    cout << "delta == " << delta << endl;
	  }
	  else {  // improvement == Good, Accurate, or NegaCurve and no restriction on increasing apply
	    cout << "Take step and increase delta, if there's room." << endl;
	    lambda = adjustLambda(lambda, 1.0, lambdaMax);
	    delta  = adjustDelta(delta, lambda, deltaMin, deltaMax);
	  }
	}
      }
      //cout << "Recompute hookstep ? " << (recompute_hookstep ? "yes" : "no") << endl;

      if (recompute_hookstep)
	continue;
      else
	break;
    }

    cout << "Taking best step and continuing Newton iteration." <<endl;
    x += dx;
    dxnorm = L2Norm(dx);
    xnorm  = L2Norm(x);

    fx = fS;

    cout << "L2Norm(f(x)) == " << L2Norm(fx) << endl;
    cout << "rx           == " << rx << endl;
    cout << "L2Norm(sN)   == " << L2Norm(sN) << endl;
    cout << "L2Norm(sH)   == " << L2Norm(sH) << endl;
    cout << "L2Norm(dx)   == " << dxnorm << endl;

  }
  unscale(x,xscale);
  return x;
}

ostream& operator<<(ostream& os, HookstepPhase p) {
  if      (p == ReducingDelta)   os << "ReducingDelta";
  else if (p == IncreasingDelta) os << "IncreasingDelta";
  else                           os << "Finished";
  return os;
}
ostream& operator<<(ostream& os, ResidualImprovement i) {
  if      (i == Unacceptable) os << "Unacceptable";
  else if (i == Poor)         os << "Poor";
  else if (i == Ok)           os << "Ok";
  else if (i == Good)         os << "Good";
  else if (i == Accurate)     os << "Accurate";
  else                        os << "NegativeCurvature";
  return os;
 }

void rescale(ColumnVector& x, const ColumnVector& xscale) {
  if (xscale.length()  != 0)
    for (int i=0; i<x.length(); ++i)
      x(i) /= xscale(i);
}
void unscale(ColumnVector& x, const ColumnVector& xscale) {
  if (xscale.length()  != 0)
    for (int i=0; i<x.length(); ++i)
      x(i) *= xscale(i);
}

Real adjustDelta(Real delta, Real deltaRate, Real deltaMin, Real deltaMax) {
  cout << "  old delta == " << delta << endl;

  // Check to see if we're already running with delta == deltaMin.
  // If so, and if deltaRate calls for furhter reduction, search has
  // ground to a halt. Error and exit.
  if (delta <= 1.00001*deltaMin && deltaRate < 1.0) {
    cout << "    delta == " << delta << endl;
    cout << "delta new == " << delta*deltaRate << endl;
    cout << "delta min == " << deltaMin << endl;
    cout << "Sorry, can't go below deltaMin. Exiting." << endl;
    exit(1);
  }

  delta *= deltaRate;

  if (delta < deltaMin) {
    delta = deltaMin;
    cout << "delta bottomed out at deltaMin" << endl;
  }
  else if (delta > deltaMax) {
    delta = deltaMax;
    cout << "delta topped out at deltaMax" << endl;
  }
  cout << "  new delta == " << delta << endl;
  return delta;
}

Real adjustLambda(Real lambda, Real lambdaMin, Real lambdaMax) {
  if (lambda < lambdaMin) {
    cout << "lambda == " << lambda << " is too small. Resetting to " << endl;
    lambda = lambdaMin;
    cout << "lambda == " << lambda << endl;
  }
  else if (lambda < lambdaMin) {
    cout << "lambda == " << lambda << " is too large. Resetting to " << endl;
    lambda = lambdaMax;
    cout << "lambda == " << lambda << endl;
  }
  return lambda;
}


void solve(const Matrix& Ut, const DiagMatrix& D, const Matrix& V, const ColumnVector& b,
	   ColumnVector& x) {

  ColumnVector bh = Ut*b;
  const Real eps = 1e-12;
  for (int i=0; i<D.rows(); ++i) {
    if (abs(D(i,i)) > eps)
      bh(i) *= 1.0/D(i,i);
    else
      bh(i)  = 0.0;
  }
  x = V*bh;
}

Real L2IP(const ColumnVector& u, const ColumnVector& v) {
  if (u.length() != v.length()) {
    cerr << "error in L2IP(ColumnVector, ColumnVector) : vector length mismatch" << endl;
    exit(1);
  }
  Real sum = 0.0;
  for (int i=0; i<u.length(); ++i)
    sum += u(i) * v(i);
  return sum;
}

Real bisectSearch(R2Rfunc f, Real a, Real b, Real feps, int maxsteps) {
  Real fa = (*f)(a);
  Real fb = (*f)(b);
  if (fa*fb > 0)
    cferror("bisectSearch(a,b,fn,xn) : a and b don't bracket a zero");

  Real c = 0.5*(a+b)/2;
  Real fc;

  for (int n=0; n<maxsteps; ++n) {
    c = 0.5*(a+b);
    fc = (*f)(c);
    if (abs(fc) < feps)
      break;
    if (fc*fa > 0) {
      a = c;
      fa = fc;
    }
    else {
      b = c;
      fb = fc;
    }
  }
  return c;
}


// The following is meant as a guide to the structure and notation
// of the above code. See Divakar Viswanath, "Recurrent motions
// within plane Couette turbulence", J. Fluid Mech. 580 (2007),
// http://arxiv.org/abs/physics/0604062 for a higher-level presentation
// of the algorithm.
//
// I've biased the notation towards conventions from fluids and from
// Predrag Cvitanivic's ChaosBook, and away from Dennis & Schnabel and
// Trefethen. That is, f stands for the finite-time map of plane Couette
// dynamics, not the residual being minimized.
//
// Let
//  G(x) = 0 : the equation we're solving
//  r(x,t) = 1/2 L2Norm2(G(x)),
//
// We want to find zeros of G(x) or equiv minimum of r(x).
// Note that the residual is reported to the user in the form
//   residual(x) == L2Norm(G(x)) == sqrt(2*r(x))
//
// The Newton algorithm for solving G(x) = 0 is
//
// Let x* be the solution, i.e. G(x*) = 0
// Suppose we start with an initial guess x near x*.
// Let x* = x + dx
// Then G(x*) = G(u+du)
//          0 = G(x) + DG(x) dx + O(|dx|^2)
// Drop h.o.t and solve
//   DG(x) dx = -G(x)
// Then start with new guess x' = x+dx and iterate.
//
// We estimate DG_{ij} = dG_i/dx_j = 1/eps (G_i(x + eps e_j) - G_i(x))
//
// Hookstep algorithm: If the Newton step dx does not decrease
// the residual L2Norm(G(x+dx)) sufficiently, we obtain the a smaller step
// s by minimizing || DG s - b ||^2 subject to ||s||^2 <= delta^2, rather
// solving DG s = b directly.

// Do minimization of | DG - b | via SVD of DG = U D V^T.
//   | DG s - b | = | U D V^T s - b   |
//                = | D V^T s - U^T b |
//                = | D sh - bh |
// where sh = V^T s and bh = U^T b. So we need to minimize |D sh - bh|
// subject to constraint |sh| <= delta. (Pure Newton step has solution
// s = sN = V D^{-1} Ut b, by the way.)

// From Divakar Viswanath (personal communication):
//
// Since D is diagonal D_{i,i} = d_i, the constrained minimization
// problem can be solved easily with lagrange multipliers:
// Let sh = sh(mu) where
//   sh_i = (bh_i d_i)/(d^2_i + mu)
// for the mu such that ||sh||^2 = delta^2. The value of mu can be
// found with a 1d Newton search for zeros of
//   Phi(mu) = ||sh(mu)||^2 - delta^2
// A straight Newton search a la mu -= Phi(mu)/Phi'(mu) is suboptimal
// since Phi'(mu) -> 0 as mu -> infty with Phi''(mu) > 0 everywhere, so
// we use a slightly modified update rule for the Newton search. Please
// refer to Dennis and Schnabel regarding that. Then, given the solution
// sh(mu), we compute the hookstep solution s from
//   V^t s = sh
//   sh = V*sh;


//#endif // HAVE_OCTAVE
// End of Octave-dependent code
// ======================================================================


} // namespace channelflow
