#ifndef CHANNELFLOW_OCTUTILS_H
#define CHANNELFLOW_OCTUTILS_H

#include <sys/stat.h>
#include <unistd.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include <vector>

#include <octave/oct.h>
#include <octave/dColVector.h>
#include <octave/dMatrix.h>
#include <octave/EIG.h>

#include "channelflow/flowfield.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/quicksort.h"

namespace channelflow {


void save(const std::string& filebase, Real  T, const FieldSymmetry& tau);
void load(const std::string& filebase, Real& T, FieldSymmetry& tau);


// Check divergence, dirichlet, normality, and maybe orthogonality of basis
void basischeck(const std::vector<RealProfile>& e, bool orthocheck=false);

// Make directory and set permissions to a+rx
void mkdir(const std::string& dirname);

// Append trailing / if necess
std::string pathfix(const std::string& path);

// Convert time to string in format convenient for filenames.
// Returns 1 or 1.000 or 1.250 as appropriate. Three digits
std::string t2s(Real t, bool decimals);

// Return false, approx, or true.
std::string fuzzyless(Real x, Real eps);

std::string pwd();

std::string clip(const std::string& filename, const std::string& ext);
std::string stub(const std::string& filename, const std::string& ext);

// Produce plots of various slices, etc.
void plotfield(const FlowField& u_, const std::string& outdir,
	       const std::string& label, int xstride=1, int ystride=1,
	       int zstride=1);

void plotspectra(const FlowField& u_, const std::string& outdir,
		 const std::string& label, bool showpadding=false);

// Produce plots of various slices, etc.
void plotxavg(const FlowField& u_, const std::string& outdir,
	      const std::string& label);

// Return a translation that maximizes the s-symmetry of u.
// (i.e. return tau // that minimizes L2Norm(s(tau(u)) - tau(u))).
FieldSymmetry optimizePhase(const FlowField& u, const FieldSymmetry& s,
			    int Nsteps=10, Real residual=1e-13, Real damp=1.0,
			    bool verbose=true, Real x0=0.0, Real z0=0.0);

// Return |P(u)|^2/|u|^2 where P(u) is projection onto
// sign ==  1 : s-symmetric subspace.
// sign == -1 : s-antisymmetric subspace.
Real PuFraction(const FlowField& u, const FieldSymmetry& s, int sign);

// A simple way to get flexible command-line args for chflow programs
// Scales as N^2 with # args, so don't use for hundreds of args.

class ArgList {
public:
  typedef std::string str; // reduces line-length below

  ArgList();
  ArgList(int argc, char* argv[], const str& purpose);

  bool helpmode() const;
  bool errormode() const;
  int remaining() const;  // how many args have yet to be parsed

  // return true if the option is in the args list, false if not
  bool getflag(const str& shortopt, const str& longopt, const str& helpstr);

  bool getbool(const str& shortopt, const str& longopt, const str& helpstr);
  int  getint (const str& shortopt, const str& longopt, const str& helpstr);
  Real getreal(const str& shortopt, const str& longopt, const str& helpstr);
  str  getstr (const str& shortopt, const str& longopt, const str& helpstr);
  str  getpath(const str& shortopt, const str& longopt, const str& helpstr);

  bool getbool(const str& shortopt, const str& longopt, bool defalt,
	       const str& helpstr);
  int  getint (const str& shortopt, const str& longopt, int defalt,
	       const str& helpstr);
  Real getreal(const str& shortopt, const str& longopt, Real defalt,
	       const str& helpstr);
  str  getstr (const str& shortopt, const str& longopt, const str& defalt,
	       const str& helpstr);
  str  getpath(const str& shortopt, const str& longopt, const str& defalt,
	       const str& helpstr);

  // In the following position counts backwards from end of arglist, e.g
  // in "argv[0] arg3 arg2 arg1" the args are numbered as indicated.
  Real getreal(int position, const std::string& name, const str& helpstr);
  str  getstr (int position, const std::string& name, const str& helpstr);

  void save(const str& outdir) const;  // save command-line to file <argv[0]>.args
  void check();  // check for unrecognized options and arguments

private:
  array<str> args_;
  array<bool> used_;
  bool helpmode_;
  bool errormode_;
  void printhelp(const str& sopt, const str& lopt, const str& type,
		 const str& defalt, const str& helpstr);
  void printhelp(int position, const str& name, const str& helpstr);
};

// ====================================================================
// All Octave-dependent code goes within this ifdef

void print(const Matrix& x);
void print(const ColumnVector& x);
void save(const Matrix& A, const std::string& filebase);
void save(const ColumnVector& x, const  std::string& filebase);
void save(const ComplexColumnVector& x, const std::string& filebase);
void setToZero(Matrix& A);
void setToZero(ColumnVector& x);

Real L2Norm(const ColumnVector& x);
Real L2Dist(const ColumnVector& x,  const ColumnVector& y);
Real L2Norm2(const ColumnVector& x);
Real L2Dist2(const ColumnVector& x,  const ColumnVector& y);
Real L2IP(const ColumnVector& u, const ColumnVector& v);

void operator*=(ColumnVector& x, Real c);

Real L2Norm(const ComplexColumnVector& x);
Real L2Dist(const ComplexColumnVector& x,  const ComplexColumnVector& y);
Real L2Norm2(const ComplexColumnVector& x);
Real L2Dist2(const ComplexColumnVector& x,  const ComplexColumnVector& y);
void operator*=(ComplexColumnVector& x, Complex c);

void quicksort(ComplexColumnVector& z); // sort by norm, decreasing
void quicksort(ComplexColumnVector& Lambda, ComplexMatrix& V); // ditto

void field2vector(const FlowField& u, ColumnVector& v,
		  const std::vector<RealProfileNG>& basis);
void vector2field(const ColumnVector& v, FlowField& u,
		  const std::vector<RealProfileNG>& basis);

void field2vector(const std::vector<BasisFunc>& basis,
		  const FlowField& u, ComplexColumnVector& v);
void vector2field(const std::vector<BasisFunc>& basis,
		  const ComplexColumnVector& v, FlowField& u);


void field2vector(const FlowField& u, ColumnVector& v);

void vector2field(const ColumnVector& v, FlowField& u);


void leastsquares(const Matrix& A, ColumnVector& b, ColumnVector& x,
		  Real& residual);

void solve(const Matrix& Ut, const DiagMatrix& D, const Matrix& C,
	   const ColumnVector& b, ColumnVector& x);


// Arnoldi iteration to estimate eigenvalues of matrix A. Usage:
// Arnoldi arnoldi(N, b);
// for (int n=0; n<N; ++n) {
//   ColumnVector q  = arnoldi.testVector();
//   ColumnVector Aq = A*q; // or however else you calculate A*q
//   arnoldi.iterate(Aq);
//   ComplexColumnVector ew = arnoldi.ew(); // current estimate of eigenvalues
// }

class Arnoldi {

public:
  Arnoldi();
  Arnoldi(const ColumnVector& b, int Niterations, Real minCondition=1e-13);

  const ColumnVector& testVector() const; // get test vector q
  void iterate(const ColumnVector& Aq);   // tell Arnoldi the value of Aq

  int n() const;     // current iteration number
  int Niter() const; // total number iterations

  const ComplexColumnVector& ew(); // current estimate of eigenvals
  const ComplexMatrix& ev();       // current estimate of eigenvecs

private:
  int M_;
  int Niter_;
  int n_;     // current iteration number
  Real condition_;

  Matrix H_;
  Matrix Q_;
  Matrix Vn_;
  ColumnVector qn_;
  ColumnVector v_;

  ComplexColumnVector ew_;
  ComplexMatrix ev_;

  void eigencalc();
};


// Iterative GMRES solution to Ax=b. Usage:
// GMRES gmres(b, N);
// for (int n=0; n<N; ++n) {
//   ColumnVector q  = gmres.testVector();
//   ColumnVector Aq = A*q; // or however else you calculate A*q
//   gmres.iterate(Aq);
//   ColumnVector x = gmres.solution(); // current estimate of soln
//   cout << "krylov residual == " << gmres.residual() << endl;
// }

// Iterative solution to Ax=b. Usage:
class GMRES {
public:
  GMRES();
  GMRES(const ColumnVector& b, int Niterations, Real minCondition=1e-13);

  const ColumnVector& testVector() const;
  void iterate(const ColumnVector& A_testvec);

  const ColumnVector& solution() const;
  const ColumnVector& guess() const;
  Real residual() const;  // |Ax-b|/|b|

  int n() const;
  int Niter() const;

  Matrix Hn() const;  // Hn  = (n+1) x n submatrix of H
  Matrix Qn() const;  // Qn  = M x n     submatrix of Q
  Matrix Qn1() const; // Qn1 = M x (n+1) submatrix of Q

private:
  int M_;     // dimension of vector space
  int Niter_; // max number of iterations
  int n_;     // current iteration number
  Real condition_;

  Matrix H_;
  Matrix Q_;
  Matrix Vn_;
  ColumnVector x0_; // guess;
  ColumnVector v_;
  ColumnVector qn_;
  ColumnVector xn_;
  Real bnorm_;
  Real residual_;
};

// Iterative solution to Ax=b with initial guess x0.
// Internally, let A (x0+y) = b and solve A y = b - A x0 with GMRES
// Residual is |Ax-b|/|b|, not |Ax-b|/|b-Ax0|
class GMRESguess {
public:
  GMRESguess(int Niterations, const ColumnVector& b, Real minCondition=1e-10);
  GMRESguess(int Niterations, const ColumnVector& b, const ColumnVector& x0, const ColumnVector& Ax0,
	     Real minCondition=1e-13);

  const ColumnVector& testVector() const;
  void iterate(const ColumnVector& A_testvec);

  ColumnVector solution() const;
  Real krylovResidual() const;

  int n() const;
  int Niter() const;

private:
  GMRES gmres_;
  ColumnVector x0_;
  Real residualScale_; // |b-Ax0|/|b|
};


class HookstepParams {
public:
  HookstepParams(Real epsSearch = 1e-14,
		 Real epsDx     = 1e-7,
		 bool centdiff  = false,
		 int  Nnewton   = 20,
		 int  Nhook     = 20,
		 Real deltaInit = 0.01,
		 Real deltaMin  = 1e-12,
		 Real deltaMax  = 1e+12,
		 Real deltaFuzz = 0.01,
		 Real lambdaMin = 0.10,
		 Real lambdaMax = 2.0,
		 Real lambdaRequiredReduction = 0.5,
		 Real improvReq = 1e-03,
		 Real improvOk   = 0.10,
		 Real improvGood = 0.75,
		 Real improvAcc  = 0.10,
		 bool pausing    = false);

  Real epsSearch; // stop search if |f(x)| < epsSearch
  Real epsDx;     // scale for perturbs in finite-diff estimate of Df
  bool centdiff;  // used centered differences in estimating Df
  int  Nnewton;   // max number of Newton step
  int  Nhook;     // max number of hookstep iterations per Newton
  Real deltaInit; // initial trust region radius
  Real deltaMin;  // stop if radius of trust region gets this small
  Real deltaMax;  // don't let trust region radius get bigger than this
  Real deltaFuzz; // accept steps within (1+/-deltaFuzz)*delta
  Real lambdaMin; // minimum delta shrink rate
  Real lambdaMax; // maximum delta expansion rate
  Real lambdaRequiredReduction; // when reducing delta, reduce by at least this factor.
  Real improvReq; // reduce delta and recompute hookstep if residual is worse than this fraction of what we'd expect from gradient
  Real improvOk;  // accept step and keep same delta if improvement is better than this fraction of quadratic mode
  Real improvGood;// accept step and increase delta if improvement is better than this fraction of quadratic model
  Real improvAcc; // recompute hookstep with larger trust region if improvement is within this fraction quadratic prediction.
  ColumnVector xscale;     // rescale x during hookstep search, x -> x./xscale. Default == don't
  ColumnVector fscale;     // rescale f during hookstep search, f -> f./fscale. Default == don't
  bool pausing;
  //string logfile;          // save diagnostic search log in this file
};

enum HookstepPhase {ConstantDelta, ReducingDelta, IncreasingDelta, Finished};
enum ResidualImprovement {Unacceptable, Poor, Ok, Good, Accurate, NegaCurve};

std::ostream& operator<<(std::ostream& os, HookstepPhase p);
std::ostream& operator<<(std::ostream& os, ResidualImprovement i);

Real adjustLambda(Real lambda, Real lambdaMin, Real lambdaMax);
Real adjustDelta(Real delta, Real rescale, Real deltaMin, Real deltaMax);


// This enum is used for labeling diagnostic output data
enum fEvalType {fEval, DfEval, HookstepEval};

typedef ColumnVector (*Rn2Rnfunc)(const ColumnVector& x, int label, bool cacheable);

void rescale(ColumnVector& x, const ColumnVector& xscale); // x -> x./xscale
void unscale(ColumnVector& x, const ColumnVector& xscale); // x -> x.*xscale

ColumnVector hookstepSearch(Rn2Rnfunc f, const ColumnVector& x0);
ColumnVector hookstepSearch(Rn2Rnfunc f, const ColumnVector& x0, const HookstepParams& p);


typedef Real (*R2Rfunc)(Real x);

Real bisectSearch(R2Rfunc f, Real a, Real b, Real feps=1e-14, int maxsteps=50);


} // namespace channelflow

#endif // OCTAVE_UTILITIES_H
